import 'package:dartz/dartz.dart';
import 'package:flutterhole/core/models/exceptions.dart';
import 'package:flutterhole/core/models/failures.dart';
import 'package:flutterhole/dependency_injection.dart';
import 'package:flutterhole/features/pihole_api/data/datasources/api_data_source.dart';
import 'package:flutterhole/features/pihole_api/data/models/dns_query_type.dart';
import 'package:flutterhole/features/pihole_api/data/models/forward_destinations.dart';
import 'package:flutterhole/features/pihole_api/data/models/over_time_data.dart';
import 'package:flutterhole/features/pihole_api/data/models/summary.dart';
import 'package:flutterhole/features/pihole_api/data/models/top_items.dart';
import 'package:flutterhole/features/pihole_api/data/models/top_sources.dart';
import 'package:flutterhole/features/pihole_api/data/repositories/api_repository.dart';
import 'package:flutterhole/features/settings/data/models/pihole_settings.dart';
import 'package:injectable/injectable.dart';

@prod
@singleton
@RegisterAs(ApiRepository)
class ApiRepositoryImpl implements ApiRepository {
  ApiRepositoryImpl([ApiDataSource apiDataSource])
      : _apiDataSource = apiDataSource ?? getIt<ApiDataSource>();

  final ApiDataSource _apiDataSource;

  /// Wrapper for accessing simple [ApiDataSource] methods.
  Future<Either<Failure, T>> fetchOrFailure<T>(
    PiholeSettings settings,
    Function dataSourceMethod,
    String description,
  ) async {
    try {
      final T result = await dataSourceMethod(settings);
      return Right(result);
    } on PiException catch (e) {
      return Left(Failure('$description failed', e));
    }
  }

  @override
  Future<Either<Failure, SummaryModel>> fetchSummary(
          PiholeSettings settings) async =>
      fetchOrFailure<SummaryModel>(
        settings,
        _apiDataSource.fetchSummary,
        'fetchSummary',
      );

  @override
  Future<Either<Failure, OverTimeData>> fetchQueriesOverTime(
          PiholeSettings settings) async =>
      fetchOrFailure<OverTimeData>(
        settings,
        _apiDataSource.fetchQueriesOverTime,
        'fetchQueriesOverTime',
      );

  @override
  Future<Either<Failure, TopSourcesResult>> fetchTopSources(
          PiholeSettings settings) async =>
      fetchOrFailure<TopSourcesResult>(
        settings,
        _apiDataSource.fetchTopSources,
        'fetchTopSources',
      );

  @override
  Future<Either<Failure, TopItems>> fetchTopItems(
          PiholeSettings settings) async =>
      fetchOrFailure<TopItems>(
        settings,
        _apiDataSource.fetchTopItems,
        'fetchTopItems',
      );

  @override
  Future<Either<Failure, ForwardDestinationsResult>> fetchForwardDestinations(
          PiholeSettings settings) async =>
      fetchOrFailure<ForwardDestinationsResult>(
        settings,
        _apiDataSource.fetchForwardDestinations,
        'fetchForwardDestinations',
      );

  @override
  Future<Either<Failure, DnsQueryTypeResult>> fetchQueryTypes(
          PiholeSettings settings) async =>
      fetchOrFailure<DnsQueryTypeResult>(
        settings,
        _apiDataSource.fetchQueryTypes,
        'fetchQueryTypes',
      );
}
